@@ -210,6 +210,13 @@ class PipelineData(np.ndarray):
             f'fs: {self.fs}, channel: {self.channel}, shape: {self.shape}'
         return result
 
+    def add_metadata(self, key, value):
+        if self.ndim > 2:
+            for md in self.metadata:
+                md[key] = value
+        else:
+            self.metadata[key] = value
+
 
 def ensure_dim(arrays, dim):
     ndim = arrays[0].ndim
@@ -549,6 +556,10 @@ def capture_epoch(epoch_s0, epoch_samples, info, target, fs=None,
     # loop.
     accumulated_data = []
     current_s0 = epoch_s0
+
+    # Info may be shared among multiple pipeline components. Make a copy so
+    # that we don't accidentally affect other pipeline components.
+    info = info.copy()
     md = info.pop('metadata', {})
 
     while True:
@@ -1194,10 +1205,37 @@ def events_to_info(trigger_edge, base_info, target):
 
 @coroutine
 def reject_epochs(reject_threshold, mode, status_cb, valid_target):
+    '''
+    Discard epochs containing artifacts
+
+    Parameters
+    ----------
+    reject_threshold : {float, callable}
+        If callble, this will be called each time this coroutine is invoked
+        allowing threshold to be updated during long-running experiments.
+    mode : {'absolute value', 'amplitude'}
+        Specifies reject method:
+        * 'absolute value': rejected if max(abs(signal)) > reject_threshold
+        * 'amplitude': rejected if (max(s) - min(s)) > reject_threshold
+    status_cb : {None, callable}
+        If provided, must be a callback that takes two arguments, the number of
+        epochs reviewed and the number of epochs rejected on that particular
+        invocation of the callback.
+    valid_target : {coroutine, callable}
+        Target to send accepted (i.e., not rejected) epochs to.
+    '''
     if mode == 'absolute value':
-        accept = lambda s: np.max(np.abs(s), axis=-1) < reject_threshold
+        accept = lambda s, th: np.max(np.abs(s), axis=-1) < th
     elif mode == 'amplitude':
-        accept = lambda s: np.ptp(s, axis=-1) < reject_threshold
+        accept = lambda s, th: np.ptp(s, axis=-1) < th
+
+    # This reduces the likelihood of name collisions (i.e., if the calling code
+    # provides a lambda that uses a variable name, it's very unlikely to be
+    # __th_cb__).
+    if not callable(reject_threshold):
+        __th_cb__ = lambda: reject_threshold
+    else:
+        __th_cb__ = reject_threshold
 
     while True:
         data = (yield)
@@ -1213,8 +1251,12 @@ def reject_epochs(reject_threshold, mode, status_cb, valid_target):
                 raise ValueError('Only one channel supported')
 
         # Check for valid epochs and send them if there are any
-        mask = accept(np.asarray(data))[:, 0]
+        th = __th_cb__()
+        mask = accept(np.asarray(data), th)[:, 0]
         valid_data = data[mask]
+        if isinstance(valid_data, PipelineData):
+            valid_data.add_metadata('reject_threshold', th)
+
         n = len(data)
         n_accept = len(valid_data)
 
