@@ -301,6 +301,9 @@ class Cos2EnvelopeFactory(EnvelopeFactory):
         super().__init__('cosine-squared', fs, duration, rise_time,
                          input_factory, start_time)
 
+    def max_amplitude(self):
+        return self.input_factory.max_amplitude()
+
 
 ################################################################################
 # SAM envelope
@@ -350,8 +353,7 @@ def sam_envelope(offset, samples, fs, depth, fm, delay, equalize):
 
 class SAMEnvelopeFactory(Modulator):
 
-    def __init__(self, fs, depth, fm, delay, direction, calibration,
-                 input_factory):
+    def __init__(self, fs, depth, fm, delay, direction, input_factory):
         vars(self).update(locals())
         self.eq_phase = sam_eq_phase(delay, depth, direction)
         self.eq_power = sam_eq_power(depth)
@@ -466,7 +468,10 @@ def _calculate_bandlimited_noise_iir(fs, calibration, fl, fh):
 
 class BandlimitedNoiseFactory(Carrier):
     '''
-    Factory for generating continuous bandlimited noise
+    Factory for generating continuous bandlimited noise using IIR filters.
+    Equalization requires a calibration that was generated using Golay.
+
+    See BandlimitedFIRNoiseFactory as an alternative.
     '''
     def __init__(self, fs, seed, level, fl, fh, filter_rolloff,
                  passband_attenuation, stopband_attenuation, equalize=False,
@@ -546,6 +551,100 @@ def bandlimited_noise(fs, level, fl, fh, duration, filter_rolloff=1,
     return factory.next(samples)
 
 
+################################################################################
+# Bandlimited FIR noise factory
+################################################################################
+class BandlimitedFIRNoiseFactory(Carrier):
+    '''
+    Factory for generating continuous shaped noise using FIR filters.
+
+    This is similar to shaped noise, but with simpler inputs if all you want is
+    bandlimited noise (i.e., no requirement to generate the dictionary of
+    gains).
+    '''
+    def __init__(self, fs, fl, fh, level, ntaps=1001, window='hann',
+                 polarity=1, seed=None, equalize=False, calibration=None):
+        vars(self).update(locals())
+
+        if calibration is None:
+            raise NotImplemented
+
+        # Calculate the gains for the shaped noise
+        if equalize:
+            freq = np.arange(fl, fh + 1)
+            sf = calibration.get_sf(freq, level)
+        else:
+            freq = np.array([fl, fh])
+            sf = calibration.get_mean_sf(fl, fh, level)
+            sf = np.full_like(freq, fill_value=sf)
+
+        freq = np.concatenate(([0, fl / 1.1], freq, [fh * 1.1, fs / 2]))
+        sf = np.pad(sf, 2)
+
+        self.taps = signal.firwin2(ntaps, freq=freq, gain=sf, window=window, fs=fs)
+        self.initial_zi = signal.lfilter_zi(self.taps, [1])
+
+        # The RMS value of noise drawn from a uniform distribution is
+        # amplitude/sqrt(3). By setting the low and high to sqrt(3) and
+        # multiplying by the scaling factors, we can ensure that the noise is
+        # initially generated with the desired RMS.
+        self.scale = np.sqrt(3)
+
+        self.reset()
+
+    def max_amplitude(self):
+        return np.abs(self.taps).sum()
+
+    def reset(self):
+        self.zi = self.initial_zi
+        self.state = np.random.RandomState(self.seed)
+        self.next(len(self.initial_zi + 1))
+
+    def next(self, samples):
+        waveform = self.state.uniform(low=-self.scale, high=self.scale, size=samples)
+        waveform, self.zi = signal.lfilter(self.taps, [1], waveform, zi=self.zi)
+        return waveform * self.polarity
+
+
+def bandlimited_fir_noise(fs, level, fl, fh, duration, ntaps=10001,
+                          window='hann', polarity=1, seed=1,
+                          calibration=None, equalize=True):
+    '''
+    Generate shaped noise using `scipy.signal.firwin2`.
+
+    Parameters
+    ----------
+    fs : float
+        Sampling rate
+    level : float
+        Level in units of calibration. If no calibration is provided, noise
+        will be scaled such that `rms(noise) == level`.
+    fl : float
+        Lower bound of noise band (Hz).
+    fh : float
+        Upper bound of noise band (Hz).
+    ntaps : int
+        Number of taps to use for filter calculation. See
+        `scipy.signal.firwin2` for hints on choosing a reasonable value.
+    window : {string, (string, float), float, None}
+        Window function to use. See `window` parameter of
+        `scipy.signal.firwin2` for additional details on acceptable values.
+    polarity : {-1, 1}
+        Polarity of noise. Useful if you need to present two trials in inverted
+        polarity to cancel out electrical artifacts.
+    seed : int
+        Seed to use for random number generator.
+    calibration : {BaseCalibration, None}
+        Instance of a `psiaudio.calibration.BaseCalibration` or subclass
+        thereof. Used to determine scaling factor for noise amplitude.
+    '''
+    args = locals()
+    args.pop('duration')
+    factory = BandlimitedFIRNoiseFactory(**args)
+    samples = int(round(duration * fs))
+    return factory.next(samples)
+
+
 ################################################################################
 # Shaped noise
 ################################################################################
@@ -561,7 +660,7 @@ class ShapedNoiseFactory(Carrier):
     '''
     Factory for generating continuous shaped noise using FIR filters.
     '''
-    def __init__(self, fs, level, gains, ntaps=1001, window='hanning',
+    def __init__(self, fs, level, gains, ntaps=1001, window='hann',
                  polarity=1, seed=None, calibration=None):
         vars(self).update(locals())
 
@@ -593,7 +692,90 @@ class ShapedNoiseFactory(Carrier):
         return waveform * self.polarity
 
 
-def shaped_noise(fs, level, gains, duration, ntaps=10001, window='hanning',
+def shaped_noise(fs, level, gains, duration, ntaps=10001, window='hann',
+                 polarity=1, seed=1, calibration=None):
+    '''
+    Generate shaped noise using `scipy.signal.firwin2`.
+
+    Parameters
+    ----------
+    fs : float
+        Sampling rate
+    level : float
+        Level in units of calibration. If no calibration is provided, noise
+        will be scaled such that `rms(noise) == level`.
+    gains : dict
+        Dictionary mapping frequency breakpoints (Hz) to gain (dB).
+    ntaps : int
+        Number of taps to use for filter calculation. See
+        `scipy.signal.firwin2` for hints on choosing a reasonable value.
+    window : {string, (string, float), float, None}
+        Window function to use. See `window` parameter of
+        `scipy.signal.firwin2` for additional details on acceptable values.
+    polarity : {-1, 1}
+        Polarity of noise. Useful if you need to present two trials in inverted
+        polarity to cancel out electrical artifacts.
+    seed : int
+        Seed to use for random number generator.
+    calibration : {BaseCalibration, None}
+        Instance of a `psiaudio.calibration.BaseCalibration` or subclass
+        thereof. Used to determine scaling factor for noise amplitude.
+    '''
+    args = locals()
+    args.pop('duration')
+    factory = ShapedNoiseFactory(**args)
+    samples = int(round(duration * fs))
+    return factory.next(samples)
+
+
+################################################################################
+# Shaped noise
+################################################################################
+def _calculate_firwin2_taps(gains, fs, window, ntaps):
+    freqs = list(gains.keys())
+    gains = util.dbi(list(gains.values()))
+    taps = signal.firwin2(ntaps, freqs, gains, fs=fs, window=window)
+    initial_zi = signal.lfilter_zi(taps, [1])
+    return taps, initial_zi
+
+
+class ShapedNoiseFactory(Carrier):
+    '''
+    Factory for generating continuous shaped noise using FIR filters.
+    '''
+    def __init__(self, fs, level, gains, ntaps=1001, window='hann',
+                 polarity=1, seed=None, calibration=None):
+        vars(self).update(locals())
+
+        self.taps, self.initial_zi = _calculate_firwin2_taps(gains, fs, window, ntaps)
+        self.sf = level if calibration is None else calibration.get_mean_sf(0, fs/2, level)
+
+        # Calculate how much the filter attenuates a *broadband* (i.e., white
+        # noise) signal. This calculation is obviously not accurate for other
+        # types of signals.
+        w, h = signal.freqz(self.taps, fs=fs)
+        h_mean = np.mean(np.abs(h) ** 2) ** 0.5
+        self.filter_sf = 1 / h_mean
+
+        # The RMS value of noise drawn from a uniform distribution is
+        # amplitude/sqrt(3). By setting the low and high to sqrt(3) and
+        # multiplying by the scaling factors, we can ensure that the noise is
+        # initially generated with the desired RMS.
+        self.scale = np.sqrt(3) * self.filter_sf * self.sf
+        self.reset()
+
+    def reset(self):
+        self.zi = self.initial_zi
+        self.state = np.random.RandomState(self.seed)
+        self.next(len(self.initial_zi + 1))
+
+    def next(self, samples):
+        waveform = self.state.uniform(low=-self.scale, high=self.scale, size=samples)
+        waveform, self.zi = signal.lfilter(self.taps, [1], waveform, zi=self.zi)
+        return waveform * self.polarity
+
+
+def shaped_noise(fs, level, gains, duration, ntaps=10001, window='hann',
                  polarity=1, seed=1, calibration=None):
     '''
     Generate shaped noise using `scipy.signal.firwin2`.
@@ -669,6 +851,86 @@ class ToneFactory(Carrier):
         return waveform
 
 
+################################################################################
+# SAMTone
+################################################################################
+def sam_tone(fs, fc, fm, level, depth=1, phase=0, polarity=1, calibration=None,
+             samples='auto', offset=0, duration=None, eq_power=True,
+             equalize=True):
+    '''
+    Generates a SAM tone
+
+    Unlike the alternate approach of combining a SAM envelope with a tone
+    carrier, this is specially designed to handle speakers with nonlinear
+    outputs as a function of frequency by adjusting the levels of the harmonics
+    accordingly.
+
+    Parameters
+    ----------
+    eq_power : bool
+        If True, compensate for modualtion depth so the overall RMS is the same
+        when varying modulation depth.  This is useful when doing AM detection
+        studies as it minimizes loudness cues.
+    equalize : bool
+        If True and the calibration is provided, ensure that the sidebands
+        (fc-fm, fc+fm) are adjusted to reflect the actual output of the speaker
+        at those frequencies. If not provided, the calibration at the carrier
+        frequency is used for both the carrier and modulation sidebands.
+
+    # TODO
+    '''
+    frequencies = fc + fm * np.arange(-1, 2)
+
+    if calibration is not None:
+        if equalize:
+            sf = calibration.get_sf(frequencies, level)
+        else:
+            sf = calibration.get_sf(fc, level)
+    else:
+        sf = level
+
+    sf = sf * np.array([0.25, 0.5, 0.25])
+    if eq_power:
+        sf /= sam_eq_power(depth)
+
+    if samples == 'auto':
+        if duration is None:
+            raise ValueError('Must provide either duration or samples')
+        samples = int(round(duration * fs))
+    elif duration is not None:
+        raise ValueError('Cannot specify duration if samples is provided')
+
+    # Since the scaling factor is based on Vrms, we need to convert this to the
+    # peak-to-peak scaling factor.
+    t = (np.arange(samples, dtype=np.double) + offset)/fs
+    s = polarity * sf[..., np.newaxis] * np.sqrt(2) * \
+        np.cos(2 * np.pi * t * frequencies[..., np.newaxis] + phase)
+    return np.sum(s, axis=0)
+
+
+class SAMToneFactory(Carrier):
+
+    def __init__(self, fs, fc, fm, level, depth=1, phase=0, polarity=1,
+                 eq_power=True, equalize=True, calibration=None):
+        vars(self).update(locals())
+        self.reset()
+
+    def reset(self):
+        self.offset = 0
+
+    def next(self, samples):
+        # Note. At least for 5 msec tones it's faster to just compute the array
+        # rather than cache the result.
+        waveform = sam_tone(
+            fs=self.fs, fc=self.fc, fm=self.fm, level=self.level,
+            depth=self.depth, phase=self.phase, polarity=self.polarity,
+            eq_power=self.eq_power, equalize=self.equalize,
+            calibration=self.calibration, offset=self.offset, samples=samples
+        )
+        self.offset += samples
+        return waveform
+
+
 ################################################################################
 # Silence
 ################################################################################
@@ -737,8 +999,11 @@ class RepeatFactory(FixedWaveform):
         self.waveform = repeat(waveform, self.fs, self.n, self.skip_n,
                                self.rate, self.delay)
 
+    def max_amplitude(self):
+        return self.input_factory.max_amplitude()
 
-def repeat(waveform, fs, n, skip_n, rate, delay=0):
+
+def repeat(waveform, fs, n, skip_n, rate, delay):
     s_period = int(round(fs / rate))
     s_waveform = len(waveform)
     s_delay = int(round(fs * delay))
@@ -759,30 +1024,61 @@ def repeat(waveform, fs, n, skip_n, rate, delay=0):
 ################################################################################
 # Chirp
 ################################################################################
+def chirp(fs, start_frequency, end_frequency, duration, level,
+          calibration=None, window='boxcar', equalize=False):
+    '''
+    Notes
+    -----
+    Windowing algorithm was implemented as described in Neumann et al., 1994 to
+    enable implementation of the Hann windowed chirp in the middel ear acoustic
+    reflex assay described by Valero et al., 2016.
+    '''
+
+    # Compute the window and get the normalized integral. This is used to
+    # adjust the instantaneous frequency so that we "dwell" longer on
+    # frequencies that are attenuated at the boundaries of the window.
+    n = int(fs * duration)
+    w = signal.get_window(window, n)
+    wi_norm = np.cumsum(w ** 2) / np.sum(w ** 2)
+    ifreq = wi_norm * (end_frequency - start_frequency) + start_frequency
+
+    # Now, we calculate the phase (the time integral of frequency). Divide by
+    # fs is equivalent to multiplying by the period (i.e., dt).
+    phase = np.cumsum(ifreq) / fs
+
+    # Figure out scaling factor. If no calibration is provided, assume that
+    # level is specified in Vrms.
+    if calibration is None:
+        if equalize:
+            raise ValueError('Cannot equalize signal without calibration')
+        sf = level
+    else:
+        if not equalize:
+            sf = calibration.get_mean_sf(start_frequency, end_frequency, level)
+        else:
+            sf = calibration.get_sf(ifreq, level)
+
+    # We need to normalize the window so that it has a RMS of 1. Then, we
+    # multiply by the square root of 2 since we are using the sin function
+    # (e.g. Vpeak = np.sqrt(2) * Vrms). Finally, multiply by the scaling factor
+    # that gives us our desired level.
+    w /= util.rms(w)
+    return np.sqrt(2) * sf * w * np.sin(2 * np.pi * phase)
+
+
 class ChirpFactory(FixedWaveform):
 
     def __init__(self, fs, start_frequency, end_frequency, duration, level,
-                 calibration):
-
-        vars(self).update(locals())
-
-        f0 = start_frequency
-        f1 = end_frequency
-
-        n = int(fs*duration)
-        t = np.arange(n, dtype=np.double) / fs
-        k = (end_frequency-start_frequency)/duration
-
-        # Compute instantaneous frequency, which can be used to compute the
-        # instantaneous scaling factor for each timepoint (thereby compensating
-        # for nonlinearities in the output).
-        ifreq = t*k + start_frequency
-        sf = calibration.get_sf(ifreq, level)*np.sqrt(2)
-
-        # Now, compute the chirp
-        self.waveform = sf*np.sin(2*np.pi*(start_frequency*t + k/2 * t**2))
+                 calibration, window='boxcar', equalize=False):
+        kwargs = locals()
+        kwargs.pop('self')
+        vars(self).update(kwargs)
+        self.waveform = chirp(**kwargs)
         self.reset()
 
+    def max_amplitude(self):
+        return np.abs(self.waveform).max()
+
 
 ################################################################################
 # Click
@@ -800,7 +1096,8 @@ class ClickFactory(FixedWaveform):
 ################################################################################
 # Bandlimited Click
 ################################################################################
-def bandlimited_click(fs, flb, fub, window=0.1, level=1):
+def bandlimited_click(fs, flb, fub, window=0.1, level=1, calibration=None,
+                      equalize=False):
     '''
     Generate bandlimited click.
 
@@ -809,17 +1106,32 @@ def bandlimited_click(fs, flb, fub, window=0.1, level=1):
 
     The click waveform will be symmetric around the center of the window.
     '''
+
     n_window = int(round(window * fs))
     n = int(round(fs))
     freq = np.fft.rfftfreq(n, d=1/fs)
     psd = np.zeros_like(freq)
     m = (freq >= flb) & (freq < fub)
     psd[m] = 1
+
+    if calibration is None:
+        if equalize:
+            raise ValueError('Cannot equalize signal without calibration')
+        sf = level
+    else:
+        if not equalize:
+            sf = calibration.get_sf(freq[m], level)
+            psd[m] = psd[m] * sf / sf.mean()
+        else:
+            sf = calibration.get_sf(freq[m], level)
+
+    sf = util.dbi(util.db(sf).mean())
+
     csd = psd * np.exp(-1j * freq * 2 * np.pi * 0.5)
-    waveform = np.fft.irfft(csd)
+    waveform = util.csd_to_signal(csd)
+    waveform = waveform / waveform.ptp() * sf
     lb = int(round(n / 2 - n_window / 2))
     waveform = waveform[lb:lb+n_window]
-    waveform /= waveform.ptp()
     return waveform
 
 
@@ -827,8 +1139,8 @@ class BandlimitedClickFactory(FixedWaveform):
 
     def __init__(self, fs, flb, fub, window, level, calibration):
         vars(self).update(locals())
-        sf = calibration.get_sf(0, level)
-        self.waveform = bandlimited_click(fs, flb, fub, window)
+        self.waveform = bandlimited_click(fs, flb, fub, window, level=level,
+                                          calibration=calibration)
         self.reset()
 
 
@@ -836,17 +1148,45 @@ class BandlimitedClickFactory(FixedWaveform):
 # Wavfiles
 ################################################################################
 @fast_cache
-def load_wav(fs, filename, level, calibration):
+def load_wav(fs, filename, level, calibration, normalization='pe'):
+    '''
+    Load wav file, scale, and resample
+
+    Parameters
+    ----------
+    fs : float
+        Desired sampling rate for wav file. If wav file sampling rate is
+        different, it will be resampled to the correct sampling rate using a
+        FFT-based resampling algorithm.
+    filename : {str, Path}
+        Path to wav file
+    level : float
+        Level to present wav files at. If normalization is `'pe'`, level will
+        be in units of peSPL (assuming calibration is in units of SPL). If
+        normalization is in `'rms'`, level will be dB SPL RMS.
+    calibration : instance of Calibration
+        Used to scale waveform to appropriate peSPL. If not provided,
+        waveform is not scaled.
+    normalization : {'pe', 'rms'}
+        Method for rescaling waveform. If `'pe'`, rescales to peak-equivalent
+        so the max value of the waveform matches the target level. If `'rms'`,
+        rescales so that the RMS value of the waveform matches the target
+        level.
+    '''
     log.warning('Loading %s', filename)
     file_fs, waveform = wavfile.read(filename, mmap=True)
-
     # Rescale to range -1.0 to 1.0
     if waveform.dtype != np.float32:
         ii = np.iinfo(waveform.dtype)
         waveform = waveform.astype(np.float32)
         waveform = (waveform - ii.min) / (ii.max - ii.min) * 2 - 1
 
-    waveform = waveform / waveform.max()
+    if normalization == 'pe':
+        waveform = waveform / waveform.max()
+    elif normalization == 'rms':
+        waveform = waveform / util.rms(waveform)
+    else:
+        raise ValueError(f'Unrecognized normalization: {normalization}')
 
     if calibration is not None:
         sf = calibration.get_sf(1e3, level)
@@ -862,33 +1202,74 @@ def load_wav(fs, filename, level, calibration):
 
 class WavFileFactory(FixedWaveform):
 
-    def __init__(self, fs, filename, level=None, calibration=None):
+    def __init__(self, fs, filename, level=None, calibration=None,
+                 normalization='pe'):
         self.fs = fs
         self.filename = filename
         self.level = level
         self.calibration = calibration
+        self.normalization = normalization
         self.reset()
 
     @property
     def waveform(self):
-        return load_wav(self.fs, self.filename, self.level, self.calibration)
+        return load_wav(self.fs, self.filename, self.level, self.calibration,
+                        normalization=self.normalization)
 
 
 class WavSequenceFactory(ContinuousWaveform):
 
-    def __init__(self, fs, path, level=None, calibration=None):
+    def __init__(self, fs, path, level=None, calibration=None, duration=-1,
+                 normalization='pe'):
+        '''
+        Parameters
+        ----------
+        fs : float
+            Sampling rate of output. If wav file sampling rate is different, it
+            will be resampled to the correct sampling rate using a FFT-based
+            resampling algorithm.
+        path : {str, Path}
+            Path to directory containing wav files.
+        level : float
+            Level to present wav files at (currently peSPL due to how the
+            normalization works).
+        calibration : instance of Calibration
+            Used to scale waveform to appropriate peSPL. If not provided,
+            waveform is not scaled.
+        duration : {None, float}
+            Duration of each wav file. If None, the wav file is loaded at the
+            beginning so that durations can be established. For large
+            directories, this can slow down the startup time of the program.
+            Knowing the exact duration may be important for some downstream
+            operations. For example, epoch extraction relative to the
+            presentation time of a particular wav file; estimating the overall
+            duration of the entire wav sequence, etc.  If you don't have a need
+            for these operations and want to speed up loading of wav files, set
+            this value to -1 (the default).
+        normalization : {'pe', 'rms'}
+            Method for rescaling waveform. If `'pe'`, rescales to
+            peak-equivalent so the max value of the waveform matches the target
+            level. If `'rms'`, rescales so that the RMS value of the waveform
+            matches the target level.
+        '''
         self.wav_files = wavs_from_path(fs, path, level=level,
-                                        calibration=calibration)
+                                        calibration=calibration,
+                                        normalization=normalization)
         self.fs = fs
+        self.duration = duration
         self.reset()
 
     def reset(self):
         self.queue = queue.BlockedRandomSignalQueue(self.fs)
-        self.queue.extend(self.wav_files, np.inf)
+        metadata = [{'filename': fh.filename.name} for fh in self.wav_files]
+        self.queue.extend(self.wav_files, np.inf, duration=self.duration, metadata=metadata)
 
     def next(self, samples):
         return self.queue.pop_buffer(samples)
 
+    def connect(self, *args, **kwargs):
+        return self.queue.connect(*args, **kwargs)
+
 
 def wavs_from_path(fs, path, *args, **kwargs):
     return [WavFileFactory(fs, filename, *args, **kwargs) \
