@@ -18,7 +18,8 @@ def as_numeric(x):
 def db(target, reference=1):
     target = as_numeric(target)
     reference = as_numeric(reference)
-    return 20*np.log10(target/reference)
+    with np.errstate(divide='ignore'):
+        return 20*np.log10(target/reference)
 
 
 def dbi(db, reference=1):
@@ -30,7 +31,7 @@ def dbtopa(db):
     '''
     Convert dB SPL to Pascal
 
-    .. math:: 10^{dB/20.0}/(20\cdot10^{-6})
+    .. math:: 10^{dB/20.0}/(20\\cdot10^{-6})
 
     >>> round(dbtopa(94), 4)
     1.0024
@@ -83,10 +84,7 @@ def normalize_rms(waveform, out=None):
     return np.divide(waveform, rms(waveform), out)
 
 
-def csd(s, window=None, waveform_averages=None, detrend='linear'):
-    if waveform_averages is not None and waveform_averages != 1:
-        new_shape = (waveform_averages, -1) + s.shape[1:]
-        s = s.reshape(new_shape).mean(axis=0)
+def csd(s, window=None, detrend='linear'):
     if detrend is not None:
         s = signal.detrend(s, type=detrend, axis=-1)
     n = s.shape[-1]
@@ -100,7 +98,7 @@ def csd(s, window=None, waveform_averages=None, detrend='linear'):
 def csd_to_signal(csd):
     n = 2 * (len(csd) - 1)
     scale = 2 / n / np.sqrt(2)
-    return np.fft.irfft(csd, axis=-1) / scale
+    return np.fft.irfft(csd / scale, axis=-1)
 
 
 def _phase(csd, unwrap=True):
@@ -120,15 +118,16 @@ def phase(s, fs, window=None, waveform_averages=None, unwrap=True):
 
 
 def psd(s, fs, window=None, waveform_averages=None, trim_samples=True):
+    s = np.asarray(s)
     if waveform_averages is None:
         waveform_averages = 1
     if trim_samples:
         n = (s.shape[-1] // waveform_averages) * waveform_averages
         s = s[..., :n]
-    new_shape = (waveform_averages, -1) + s.shape[1:]
+    new_shape = s.shape[:-1] + (waveform_averages, -1)
     s = s.reshape(new_shape)
-    c = csd(s, window)
-    return np.abs(c).mean(axis=0)
+    c = csd(s, window=window)
+    return np.abs(c).mean(axis=-2)
 
 
 def psd_freq(s, fs):
@@ -147,6 +146,39 @@ def csd_df(s, fs, *args, **kw):
 
 
 def psd_df(s, fs, *args, waveform_averages=None, **kw):
+    '''
+    Compute PSD and return as a dataframe with columns indexed by frequency
+
+    Parameters
+    ----------
+    s : {array, DataFrame}
+        Data to compute PSD over. The last axis (columns for DataFrame) is
+        always assumed to be time. If array, only 2D arrays are supported (for
+        conversion into DataFrame).
+    fs : float
+        Sampling rate of data.
+    window : {None, string}
+        Applies given window to signal before computing FFT. If None, no
+        windowing is performed. Name of window must be a valid window name that
+        can be passed to `scipy.signal.get_window`.
+    waveform_averages : {None, int}
+        Number of chunks to segment time data into before computing PSD.
+        Averaging is done after computing the PSD.
+    trim_samples : bool
+        If true, remove excess time samples so that waveform can be split into
+        `waveform_averages` segments of equal size. If False and the number of
+        timepoints is not an integer multiple of `waveform_averages`, an error
+        will be raised.
+
+    Additional arguments are passed to :func:`~psiaudio.util.psd`.
+
+    Returns
+    -------
+    psd : pd.DataFrame
+        DataFrame with frequency as columns. If ``s`` was a DataFrame, the
+        original index of ``s`` will be preserved. If ``s`` was an array, psd
+        will have a simple integer index.
+    '''
     p = psd(s, fs, *args, waveform_averages=waveform_averages, **kw)
     n = s.shape[-1]
     if waveform_averages is not None:
@@ -287,30 +319,64 @@ def analyze_tone(waveforms, frequency, fs, mic_gain, trim=0, thd_harmonics=3):
     }
 
 
-def spectrum_to_band_level(spectrum_db, flb, fub):
+def spectrum_to_band_level(spectrum_db, n):
+    '''
+    Convert spectrum level to overall level
+
+    Parameters
+    ----------
+    spectrum_db : float
+        Spectrum level in dB re. reference (e.g., dB SPL).
+    n : int
+        Number of bands. If you are trying to compare this to an FFT with 1 Hz
+        resolution, this is equal to the bandwidth (in Hz). If you are
+        comparing this to an FFT with a bin size of 500 Hz, then this is equal
+        to the bandwidth / 500.
+
+    Returns
+    -------
+    band_db : float
+        Band level in same units as spectrum level.
+    '''
+    return spectrum_db + 10 * np.log10(n)
+
+
+def band_to_spectrum_level(band_db, n):
     '''
     Convert overall band level to spectrum level
+
+    Parameters
+    ----------
+    band_db : float
+        Band level in dB re. reference (e.g., dB SPL).
+    n : int
+        Number of bands. If you are trying to compare this to an FFT with 1 Hz
+        resolution, this is equal to the bandwidth (in Hz). If you are
+        comparing this to an FFT with a bin size of 500 Hz, then this is equal
+        to the bandwidth / 500.
+
+    Returns
+    -------
+    spectrum_db : float
+        Spectrum level in same units as band level.
     '''
-    return spectrum_db + 10 * np.log10(fub - flb)
+    return band_db - 10 * np.log10(n)
 
 
-def band_to_spectrum_level(band_db, flb, fub):
-    '''
-    Convert overall band level to spectrum level
-    '''
-    return band_db - 10 * np.log10(fub - flb)
-
-
-def rms(s, detrend=False):
+def rms(s, detrend=False, axis=-1):
     if detrend:
-        s = signal.detrend(s, axis=-1)
-    return np.mean(s**2, axis=-1)**0.5
+        s = signal.detrend(s, axis=axis)
+    return np.mean(s**2, axis=axis)**0.5
 
 
 def rms_rfft(x):
     return np.sqrt(np.sum(np.abs(x) ** 2))
 
 
+def rms_rfft_db(x, *args, **kw):
+    return db(rms_rfft(dbi(x, *args, **kw)))
+
+
 def golay_pair(n=15):
     '''
     Generate pair of Golay sequences
@@ -697,7 +763,39 @@ def resample_poly(waveform, fs, target_fs):
     return signal.resample_poly(waveform, up, down, axis=-1)
 
 
-def psd_bootstrap(x, fs, n_draw=400, n_bootstrap=100, rng=None, window=None):
+def psd_bootstrap_vec(x, fs, n_draw=400, n_bootstrap=100, rng=None, window=None):
+    '''
+    Calculate the normalized PSD across trials using a boostrapping algorithm
+
+    To estmate the noise floor, the CSD of each trial is computed and then the
+    phases of the CSD are randomized.
+
+    Parameters
+    ----------
+    x : array
+        Signal to compute PSD noise floor over. Must be trial x time.
+    fs : float
+        Sampling rate of signal
+    n_draw : int
+        Number of trials to draw on each bootstrap cycle.
+    n_bootstrap : int
+        Number of bootstrap cycles.
+    rng : instance of RandomState
+        If provided, this will be used to drive the bootstrapping algorithm.
+        Useful when you need to "freeze" results (e.g., for publication).
+    window : {None, string}
+        Type of window to use when calculating bootstrapped PSD.
+
+    Result
+    ------
+    psd_bs : DataFrame
+        Pandas DataFrame indexed by frequency. Columns include `psd_nf`, the
+        noise floor as estimated by the bootstrapping algorithm.
+
+    Notes
+    -----
+    TODO: Add citation (Bharadwaj).
+    '''
     if rng is None:
         rng = np.random.RandomState()
 
@@ -730,6 +828,73 @@ def psd_bootstrap(x, fs, n_draw=400, n_bootstrap=100, rng=None, window=None):
     )
 
 
+def psd_bootstrap_loop(x, fs, n_draw=400, n_bootstrap=100, rng=None, window=None):
+    '''
+    Calculate the normalized PSD across trials using a boostrapping algorithm
+
+    To estmate the noise floor, the CSD of each trial is computed and then the
+    phases of the CSD are randomized.
+
+    Parameters
+    ----------
+    x : array
+        Signal to compute PSD noise floor over. Must be trial x time.
+    fs : float
+        Sampling rate of signal
+    n_draw : int
+        Number of trials to draw on each bootstrap cycle.
+    n_bootstrap : int
+        Number of bootstrap cycles.
+    rng : instance of RandomState
+        If provided, this will be used to drive the bootstrapping algorithm.
+        Useful when you need to "freeze" results (e.g., for publication).
+    window : {None, string}
+        Type of window to use when calculating bootstrapped PSD.
+
+    Result
+    ------
+    psd_bs : DataFrame
+        Pandas DataFrame indexed by frequency. Columns include `psd_nf`, the
+        noise floor as estimated by the bootstrapping algorithm.
+
+    Notes
+    -----
+    TODO: Add citation (Bharadwaj).
+    '''
+    if rng is None:
+        rng = np.random.RandomState()
+
+    c = csd(x, window=window)
+    i = np.arange(len(c))
+
+    mean_psd_bs = []
+    mean_psd_bs_rand = []
+    plv_bs = []
+
+    for b in range(n_bootstrap):
+        c_bs = c[rng.choice(i, n_draw)]
+        random_phases = rng.uniform(0, 2 * np.pi, size=c_bs.shape)
+        c_bs_rand = np.abs(c_bs) * np.exp(-1j * random_phases)
+        mean_psd_bs.append(db(np.abs(c_bs.mean(axis=0))))
+        mean_psd_bs_rand.append(db(np.abs(c_bs_rand.mean(axis=0))))
+        angle_bs = np.angle(c_bs)
+        plv_bs.append(np.abs(np.mean(np.exp(-1j*angle_bs), axis=0)))
+
+    plv = np.mean(plv_bs, axis=0)
+    psd_nf = np.mean(mean_psd_bs_rand, axis=0)
+    psd = np.mean(mean_psd_bs, axis=0)
+    psd_norm = psd - psd_nf
+
+    return pd.DataFrame({
+            'psd_nf': psd_nf,
+            'psd': psd,
+            'psd_norm': psd_norm,
+            'plv': plv,
+        },
+        index=pd.Index(psd_freq(x, fs), name='frequency'),
+    )
+
+
 ################################################################################
 # Multichannel functions
 ################################################################################
@@ -817,11 +982,15 @@ def epochs(x, pad=0):
 def smooth_epochs(epochs):
     '''
     Given a 2D array of epochs in the format [[start time, end time], ...],
-    identify and remove all overlapping epochs such that::
+    identify and remove all overlapping epochs such that:
+
         [ epoch   ]        [ epoch ]
             [ epoch ]
-    Will become::
+
+    Will become:
+
         [ epoch     ]      [ epoch ]
+
     Epochs do not need to be ordered when provided; however, they will be
     returned ordered.
     '''
@@ -894,17 +1063,18 @@ def int_to_TTL(a, width):
     represented as a single bit (0 or 1), it is wasteful to cast the TTL to an
     int32 before storing the data in a buffer.  `FromBits` combines up to 6 TTL
     channels into a single int32 word.  Since only the first 6 bits are used to
-    store the 6 TTL channels, the data can be reduced further.
+    store the 6 TTL channels, the data can be reduced further:
+
     1. Using `ShufTo8`, 24 TTL channels can be stored in a single index of a
-    serial buffer.
+       serial buffer.
     2. Using `CompTo8`, 4 consecutive samples of data from 6 TTL channels can be
-    stored in a single index of a serial buffer.
+       stored in a single index of a serial buffer.
     3. Combining `ShufTo16` and `CompTo16`, store 2 consecutive samples of data
-    from 12 TTL channels in a single index of a serial buffer.
-    Using this approach, the memory overhead and amount of data being
-    transferred has been reduced by a factor of 24.
-    This function uses Numpy's bitshift and bitmask operators, so the algorithm
-    should be pretty efficient.
+       from 12 TTL channels in a single index of a serial buffer. Using this
+       approach, the memory overhead and amount of data being transferred has
+       been reduced by a factor of 24.  This function uses Numpy's bitshift and
+       bitmask operators, so the algorithm should be pretty efficient.
+
     Parameters
     ==========
     a : array_like
@@ -912,12 +1082,13 @@ def int_to_TTL(a, width):
         The dtype (either int8, int16 or int32) of the array is used to figure
         out the size of the second dimension.  This will depend on your
         combination of `FromBits` and the shuffle/compression components.
+
     Returns
     =======
     bitfield : array
         2D boolean array repesenting the bits in little-endian order
 
-    Example (note the transpose -- didn't have time to flip around)
+    Example
     =======
     >>> int_to_TTL([4, 8, 5], width=6).T
     array([[False, False,  True, False, False, False],
