@@ -133,6 +133,15 @@ def test_pipeline_data_astype():
     assert_pipeline_data_equal((pd > 0.5), pd_th.astype('bool'))
 
 
+def test_negative_s0():
+    md = {'foo': 'bar'}
+    d = np.random.uniform(size=100000)
+    data = pipeline.PipelineData(d, 100e3, metadata=md, s0=-64)
+    assert data[::8].s0 == -64
+    assert data[64:].s0 == 0
+    assert data[32:].s0 == -32
+
+
 def test_pipeline_data_1d(data1d):
     fs = data1d.fs
     md = data1d.metadata.copy()
@@ -595,6 +604,7 @@ def test_reject_epochs(fs, data_fixture, detrend_mode, request):
     if data_fixture == 'data1d':
         data[s0] = 3
         queue, expected, epoch_size = queue_epochs(data)
+        expected.add_metadata('reject_threshold', 2)
         actual = pipeline.concat(feed_pipeline(cb, data), axis=-3)
         assert_pipeline_data_equal(actual, expected[[0, 2]])
     elif data_fixture == 'data2d':
@@ -606,9 +616,63 @@ def test_reject_epochs(fs, data_fixture, detrend_mode, request):
         actual = pipeline.concat(feed_pipeline(cb, data[1]), axis=-3)
         # Avoid weird numpy indexing behavior
         expected = expected[[0, 2]][:, [1]]
+        expected.add_metadata('reject_threshold', 2)
         assert_pipeline_data_equal(actual, expected)
 
 
+@pytest.mark.parametrize('data_fixture', ['data1d'])
+def test_reject_epochs_update_th(fs, data_fixture, detrend_mode, request):
+    data = request.getfixturevalue(data_fixture)
+
+    # Lambda functions will pull the value reject_threshold from the namespace.
+    th_cb = lambda: th
+
+    def cb(target):
+        nonlocal data
+        nonlocal queue
+        nonlocal epoch_size
+
+        return \
+            pipeline.extract_epochs(
+                fs=data.fs,
+                queue=queue,
+                epoch_size=epoch_size,
+                target=pipeline.reject_epochs(
+                    th_cb,
+                    'amplitude',
+                    None,
+                    target
+                ).send
+            )
+
+    queue, expected, epoch_size = queue_epochs(data)
+
+    segments = []
+    actual = []
+    cr = cb(target=actual.append)
+    o = 0
+
+    thresholds = [2, 4, 2]
+    for e, th in zip(expected, thresholds):
+        # Now, add an artifact value of 3 to each.
+        s0 = int(round(e.metadata['t0'] * e.fs))
+        e0 = s0 + e.n_time
+        i = np.random.randint(s0, e0)
+        data[i] = 3
+        cr.send(data[o:e0])
+        o = e0
+    cr.send(data[o:])
+
+    # We need to re-run the standard epoch queue using the mutated data (where
+    # we added the artifacts) so that the resulting epoch waveforms are
+    # identical to the ones used for the artifact reject.
+    _, expected, _ = queue_epochs(data)
+    expected = expected[[1]]
+    expected.add_metadata('reject_threshold', 4)
+    actual = pipeline.concat(actual, axis=-3)
+    assert_pipeline_data_equal(actual, expected)
+
+
 def test_edges():
     # Supporting function to handle segmenting PipelineData into 10 chunks and
     # then feeding each chunk into the pipeline (useful for verifying that we
